// chugl modules you'd want to use
//Machine.add( "KB.ck" );
//Machine.add( "Mouse.ck" );
//Machine.add( "FlyCam.ck" );

//GG.fullscreen();

// ------------------ some constants -----------------------//

// window size
//1024 => int WINDOW_SIZE;
// y position of waveform
//2 => float WAVEFORM_Y;
// y position of spectrum
//-2.5 => float SPECTRUM_Y;
// width of waveform and spectrum display
10 => float DISPLAY_WIDTH;
// waterfall depth
500 => int WATERFALL_DEPTH;
50 => int MOONFALL_DEPTH;
// y position of waveform
2 => float WAVEFORM_Y;

256 => int AXIS_LENGTH;
200.  => float size_X;
150.  => float size_Y;
300. => float size_Z;

GG.camera().clip( GG.camera().clipNear(), 1000);

<<< GG.camera().clipNear(), GG.camera().clipFar() >>>;

// coordinate scales
size_X / AXIS_LENGTH     => float scale_X;
size_Y                   => float scale_Y;
size_Z / WATERFALL_DEPTH => float scale_Z;

500. => float GLOBE_RADIUS;
Math.pi / 4 => float INITIAL_ROT;
-Math.pi => float TOTAL_ROT;

500. => float WAVEFORM_SCALE;
550. => float WAVEFORM_RADIUS;

// manipulate scroll speed while keeping terrain generation shape same
1. => float SCROLL_SPEED;

fun void change_scroll(float target_scroll, float duration) {
    5 => int T;
    SCROLL_SPEED => float initial_scroll;
    for (0 => int t; t <= T; t++) {
        (target_scroll * t + initial_scroll * (T - t)) / T  => SCROLL_SPEED;
        duration::second / (T + 1) => now;
    }
}


// ------------------ scene setup -------------------------- //

GG.scene() @=> GScene @ scene;   // scene reference
GG.camera().pos(@(-100, 5, -100));
GG.camera().lookAt( @(0, 10, 0));  // angle camera

GG.camera().fov(70.);

FlyCam flyCam;
flyCam.init(0);
spork ~ flyCam.selfUpdate();

0 => int playMode;
0 => int MODE_INTERACTIVE;
1 => int MODE_PLAYBACK;

// ------------------ fog for ominous effect --------------- //


// fog setup =====================

//scene.enableFog();
//@(0, 0.42, 0.58) => vec3 fogColor;
//@(0.2, 0.6, 0.8) => vec3 fogColor;
@(0., 0., 0.) => vec3 fogColor;


// important! match fog color and background color for more realistic effect
scene.fogColor(fogColor);
scene.backgroundColor(fogColor);

// oscillate fog density between [0 and 0.6]
fun void pingPongFogDensity() {
    while (true) {
        scene.fogDensity(Math.sin(now/10::second) * 0.006 + 0.007);
        GG.nextFrame() => now;
    }
}
//spork ~ pingPongFogDensity();


// --------------------------------------------------------- //
// Everything is just sndpeek.....


// window size
AXIS_LENGTH => int WINDOW_SIZE;

// make a waterfall
Waterfall waterfall --> GG.scene();

// custom GGen to render waterfall
class Waterfall extends GGen
{
    // waterfall playhead
    0 => int playhead;
    // terrain
    GMesh wfl[WATERFALL_DEPTH];
    DataTexture textures[WATERFALL_DEPTH];   // workaround for garbage collection bug 
    
    // rotational axis
    GGen connector[WATERFALL_DEPTH];
    GGen axis[WATERFALL_DEPTH];
    
    // for terrain generation
    vec3 old_terrain[AXIS_LENGTH];
    
    fun void init() {
        // Initialize terrain
        for (0 => int i; i < old_terrain.size(); i++) {
            @(scale_X * -i, 0, -1.) => old_terrain[i];
        }
        
        // setup mesh and axis
        playhead => int pos;
        for (0 => int i; i < WATERFALL_DEPTH; i++)
        {
            wfl[i] --> connector[i] --> axis[i] --> this;
            
            // set positions
            axis[i].posY(-GLOBE_RADIUS);
            axis[i].posZ(GG.camera().pos().z);
            connector[i].posY(GLOBE_RADIUS - 40.);
            
            // start with playhead-1 and go backwards
            pos--; if( pos < 0 ) WATERFALL_DEPTH-1 => pos;
            axis[pos].rot( @(0, 0, 0) );
            axis[pos].rotateX( INITIAL_ROT + TOTAL_ROT * i / axis.size() );
            
            // scale terrain (1 -> arc length of each segment)
            GLOBE_RADIUS * SCROLL_SPEED * Math.fabs(TOTAL_ROT) / WATERFALL_DEPTH => float scalar;
            axis[i].sca( @(1, 1, scalar) );
        }
    }
    init();
    
    fun void latest( vec3 new_terrain[] )
    {
        get_terrain_strip(textures[playhead], old_terrain, new_terrain) @=> GMesh newMesh;
        // shift new_terrain over
        for (0 => int i; i < old_terrain.size(); i++) {
            new_terrain[i] + @(0, 0, -1) => old_terrain[i];
        }
        
        // set into
        wfl[playhead] --< connector[playhead];
        newMesh @=> wfl[playhead];
        newMesh --> connector[playhead];
        // advance playhead
        playhead++;
        // wrap it
        WATERFALL_DEPTH %=> playhead;
        
        // position
        playhead => int pos;
        // so good
        for( 0 => int i; i < axis.size(); i++ )
        {
            // start with playhead-1 and go backwards
            pos--; if( pos < 0 ) WATERFALL_DEPTH-1 => pos;
            
            // spherical wrap
            axis[pos].rotateX( SCROLL_SPEED * TOTAL_ROT / axis.size() );
            if (i + 1 == axis.size()) {
                axis[pos].rot( @(0, 0, 0) );    // reset rotation
                axis[pos].rotateX( INITIAL_ROT );
            }
        }
        
        // scale terrain (1 -> arc length of each segment)
        GLOBE_RADIUS * SCROLL_SPEED * Math.fabs(TOTAL_ROT) / WATERFALL_DEPTH => float scalar;
        axis[pos].sca( @(1, 1, scalar) );
    }
}

// Create strip of terrain given two neighboring tiles
// terrain1 is the old one
// terrain2 is the new one
fun GMesh get_terrain_strip(DataTexture dataTexture, vec3 terrain1[], vec3 terrain2[]) {
    if (terrain1.size() != terrain2.size()) {
        <<< "size mismatch in get_terrain_strip()", "" >>>;
        return GMesh whatever;
    }
    
    // DataTexture from terrain
    //DataTexture dataTexture;
    float audioTextureData[0];     // .size() == (AXIS_LENGTH x 2) x 4; (r, g, b, a)
    attach(audioTextureData, getNewAudioTexture(terrain2));
    attach(audioTextureData, getNewAudioTexture(terrain1));
    dataTexture.data(audioTextureData, AXIS_LENGTH, 2);
    
    // All the fun stuff
    CustomGeometry myGeometry;
    vec3 all_vertices[0];
    int indices[0];
    float uvs[0];
    
    attach(all_vertices, terrain1);
    attach(all_vertices, terrain2);
    
    // Calculate midpoints
    vec3 midpoints[0];
    for (0 => int j; j < terrain1.size() - 1; j++) {
        all_vertices << (terrain1[j] + terrain1[j+1] + terrain2[j] + terrain2[j+1]) / 4;
    }
    
    // Calculate triangle meshes indices
    for (0 => int j; j < terrain1.size() - 1; j++) {
        indices << j;
        indices << terrain1.size() + j;
        indices << terrain1.size() + terrain2.size() + j;
        
        indices << j;
        indices << j+1;
        indices << terrain1.size() + terrain2.size() + j;
        
        indices << terrain1.size() + j;
        indices << terrain1.size() + j+1;
        indices << terrain1.size() + terrain2.size() + j;
        
        indices << j+1;
        indices << terrain1.size() + (j+1);
        indices << terrain1.size() + terrain2.size() + j;
    }
    
    // Calculate uvs of each vertex
    uvs.reset();
    for (0 => int i; i < terrain1.size(); i++) {
        uvs << 1. * i / (AXIS_LENGTH - 1);
        uvs << 1.;
    }
    for (0 => int i; i < terrain1.size(); i++) {
        uvs << 1. * i / (AXIS_LENGTH - 1);
        uvs << 0.;
    }
    for (0 => int i; i < terrain1.size() - 1; i++) {
        uvs << (i + 0.5) / (AXIS_LENGTH - 1);
        uvs << 0.5;
    }
    
    myGeometry.positions( all_vertices );
    myGeometry.indices( indices );
    myGeometry.uvs( uvs );

    // Create material
    FlatMaterial myMat;
    //MangoUVMaterial myMat;
    myMat.diffuseMap(dataTexture);
    myMat.polygonMode( Material.POLYGON_FILL );
    
    // Create mesh
    GMesh customMesh;
    customMesh.set(myGeometry, myMat);
    return customMesh;
}


// ---------------------------------------------------------//

// Helper function: attach midpoints to positions
fun void attach (vec3 arr1[], vec3 arr2[]) {
    for (0 => int i; i < arr2.size(); i++) {
        arr1 << arr2[i];
    }
}

fun void attach (float arr1[], float arr2[]) {
    for (0 => int i; i < arr2.size(); i++) {
        arr1 << arr2[i];
    }
}

fun void attach (float arr1[], vec4 pt) {
    arr1 << pt.x;
    arr1 << pt.y;
    arr1 << pt.z;
    arr1 << pt.w;
}

// -------------------------------------------//

// Colorful colors!!!
@(1,41,74,255) => vec4 deep_blue;
@(1,67,102,255) => vec4 medium_blue;
@(0,109,150,255) => vec4 light_blue;
@(232,242,246,255) => vec4 bubbly_white;
@(232,227,223,255) => vec4 beachy_tan;
@(160,145,150,255) => vec4 rocky_brown;
@(70,53,35,255) => vec4 dark_rocky_brown;
@(137,179,217,255) => vec4 mountainous_blue;

-size_X * 0.9 => float thres1;
-size_X * 0.6 => float thres2;
-size_X * 0.2 => float thres3;

size_Y * 0.04 => float thres4;
size_Y * 0.06 => float thres5;
size_Y * 0.03 => float thres6;

size_Y * 0.25 => float thres7;
size_Y * 0.50 => float thres8;
size_Y * 0.9 => float thres9;


fun vec4 get_color( vec3 pt ) {
    vec4 color;
    
    // Ocean color + ocean whites
    if (pt.y < thres5) {
        if (pt.x < thres1) {
            deep_blue => color;
        } else if (pt.x < thres2) {
            (deep_blue * (thres2 - pt.x) + medium_blue * (pt.x - thres1)) / (thres2 - thres1) => color;
        } else if (pt.x < thres3) {
            (medium_blue * (thres3 - pt.x) + light_blue * (pt.x - thres2)) / (thres3 - thres2) => color;
        } else {
            light_blue => color;
        }
        
        if (pt.y > thres4) {
            //rocky_brown => color;

            (rocky_brown * (thres5 - pt.y) + color * (pt.y - thres4)) / (thres5 - thres4) => color;
        } else if (pt.y > thres6) {

            (bubbly_white * (thres6 - pt.y) + color * (pt.y - thres4)) / (thres6 - thres4) => color;
        }
    } else if (pt.y < thres7) {
        dark_rocky_brown => color;
    } else if (pt.y < thres8) {
        rocky_brown => color;
    } else if (pt.y < thres9) {
        mountainous_blue => color;
    } else {
        bubbly_white => color;
    }
    
    return color;
}

fun float[] getNewAudioTexture(vec3 new_terrain[]) {
    float newAudioTextureData[0];
    for (0 => int i; i < new_terrain.size(); i++) {
        attach(newAudioTextureData, get_color(new_terrain[i]));
    }
    
    return newAudioTextureData;
}


// map FFT output to 3D positions
fun int map2spectrum( complex in[], vec3 out[] )
{
    // mapping to xyz coordinate
    int i;
    DISPLAY_WIDTH => float width;
    //for( auto s : in )
    //{
    //    // space evenly in X
    //    scale_X * -i => out[i].x;
    //    // map frequency bin magnitide in Y
    //    scale_Y * Math.sqrt( (s$polar).mag * 25 ) => out[i].y;
    //    // constant 0 for Z here
    //    scale_Z * 0 => out[i].z;
    //    // increment
    //    i++;
    //}
    for (0 => int i; i < in.size(); i++)
    {
        in[i] => complex s;
        // space evenly in X
        //scale_X * (INITIAL_ROT + FINAL_ROTATION => out[i].x;
        scale_X * -i => out[i].x;
        // map frequency bin magnitude in Y
        scale_Y * Math.sqrt( (s$polar).mag * 25 ) => out[i].y;
        // constant 0 for Z here
        0 => out[i].z;
        // increment
        i++;
    }

    
    // rest of the out vector is flat
    for (in.size() => int i; i < out.size(); i++) {
        scale_X * -i => out[i].x;
        0 => out[i].y;
        0 => out[i].z;
    }
    
    return 0;
}

// ------------------------- floating donut ??? -------------------- //

// make a waterfall
Moonfall moonfall --> GG.scene();

// custom GGen to render waterfall
class Moonfall extends GGen
{
    // waterfall playhead
    0 => int playhead;
    // terrain
    GLines wfl[MOONFALL_DEPTH];
    
    for (0 => int i; i < wfl.size(); i++) {
        wfl[i] --> this;
    }
    
    fun void latest( float samples[] )
    {
        get_waveform( wfl[playhead], samples );
        
        // advance playhead
        playhead++;
        // wrap it
        MOONFALL_DEPTH %=> playhead;
        
        // position
        playhead => int pos;
        // so good
        for( 0 => int i; i < wfl.size(); i++ )
        {
            // start with playhead-1 and go backwards
            pos--; if( pos < 0 ) MOONFALL_DEPTH-1 => pos;
            
            // away into space!
            wfl[pos].posY(i * 5);
            wfl[pos].sca( 1. * (wfl.size() - i) / wfl.size() * @(1, 1, 1) );
        }
        
        // scale terrain (1 -> arc length of each segment)
        //GLOBE_RADIUS * SCROLL_SPEED * Math.fabs(TOTAL_ROT) / WATERFALL_DEPTH => float scalar;
        //axis[pos].sca( @(1, 1, scalar) );
    }
}

fun GLines get_waveform( GLines waveform, float samples[] ) {
    // waveform renderer
    //GLines waveform; 
    waveform.mat().color( @(.4, .4, .1) );
        
    // mapping to xyz coordinate
    vec3 positions[WINDOW_SIZE];
    for(0 => int i; i < samples.size(); i++)
    {
        -2 * Math.pi * i / WINDOW_SIZE => float phase;
        Math.fabs(samples[i]) * 60 => float amplitude;
        %(amplitude, phase) => polar pol;
        pol $ complex => complex coordinates;
        
        WAVEFORM_SCALE * amplitude / (WAVEFORM_RADIUS) => float arc_angle;
        if (arc_angle > 3 * Math.acos(GLOBE_RADIUS / WAVEFORM_RADIUS)) {
            3 * Math.acos(GLOBE_RADIUS / WAVEFORM_RADIUS) => arc_angle; 
        }
        
        WAVEFORM_SCALE * coordinates.re * Math.cos(arc_angle) => positions[i].x;
        WAVEFORM_RADIUS *                 Math.cos(arc_angle) => positions[i].y;
        WAVEFORM_SCALE * coordinates.im * Math.cos(arc_angle) => positions[i].z;
    }
    
    waveform.geo().positions( positions );
    waveform.pos( @(GG.camera().posX(), -GLOBE_RADIUS, GG.camera().posZ()) );
    
    return waveform;
}



// ------------------------- AUDIO STUFF --------------------------- //

// spectrum renderer
GLines spectrum --> GG.scene(); spectrum.mat().lineWidth(1.0);
// translate down
spectrum.posY(-WAVEFORM_Y);
// color0
spectrum.mat().color( @(.4, 1, .4) );

// common audio buffer
UGen audio_buf => Flip accum => blackhole;
audio_buf => PoleZero dcbloke => FFT fft => blackhole;
audio_buf => dac;

// this one for input audio
UGen input_buf => accum;
input_buf => dcbloke;

// music narrative
global UGen narrative;
narrative.gain(0);

// accumulate samples from mic
adc => input_buf;

// set DC blocker
.95 => dcbloke.blockZero;
// set size of flip
WINDOW_SIZE => accum.size;
// set window type and size
Windowing.hann(WINDOW_SIZE) => fft.window;
// set FFT size (will automatically zero pad)
WINDOW_SIZE*2 => fft.size;
// get a reference for our window for visual tapering of the waveform
Windowing.hann(WINDOW_SIZE) @=> float window[];

// sample array
float samples[0];
// FFT response
complex response[0];

// do audio stuff
fun void doAudio()
{
    while( true )
    {
        // upchuck to process accum
        accum.upchuck();
        // get the last window size samples (waveform)
        accum.output( samples );
        // upchuck to take FFT, get magnitude reposne
        fft.upchuck();
        // get spectrum (as complex values)
        fft.spectrum( response );
        // jump by samples
        WINDOW_SIZE::samp/2 => now;
    }
}
spork ~ doAudio();



// ------------------------------------------------ //

// fps printer
fun void printFPS( dur howOften )
{
    while( true )
    {
        <<< "fps:", GG.fps() >>>;
        howOften => now;
    }
}
//spork ~ printFPS(.25::second);



vec3 new_terrain[WINDOW_SIZE];

// graphics render loop
while( true )
{
    // map to spectrum display
    map2spectrum( response, new_terrain ) @=> int output;
    
    if (output != -1) {
        waterfall.latest( new_terrain );
        moonfall.latest( samples );
    }
    
    // next graphics frame
    //GG.nextFrame() => now;
    50::ms => now;
}
