global UGen narrative;

// ----------------- HELPER FUNCTIONS ------------ //

fun void change_volume(UGen buf, float target_volume, float duration) {  
    10000 => int T;
    buf.gain() => float initial_volume;
    for (0 => int t; t <= T; t++) {
        buf.gain( (target_volume * t + initial_volume * (T - t)) / T );
        duration::second / (T + 1) => now;
    }
}


// -------------------- INTRO ------------------- //
fun void play_intro() {
    SinOsc s_1 => JCRev r_1 => narrative;
    .3 => r_1.gain;
    .2 => r_1.mix;
    
    // note number
    20 => float note;
    
    // go up to 127
    while( note < 128 )
    {
        // convert MIDI note to hz
        Std.mtof( note ) => s_1.freq;
        // turn down the volume gradually
        .5 - (note/256.0) => s_1.gain;
        
        // move up by whole step
        note + 1 => note;
        
        // advance time
        .125::second => now;
    }
    
    // turn off s
    spork ~ change_volume(s_1, 0., 1.);
    // wait a bit
    4::second => now;
}
play_intro();







// ------------------- SOME NOTES ------------------- //

// low-pass filter for the ~vibes~
Gain notes;
notes.gain(0.);
LPF lpf;
lpf.freq(2000);
lpf => notes => narrative;

fun void play_narrative_notes() {
    // scale degrees in semi-tones
    [ 0, 2, 4, 7, 9 ] @=> int f[];
    
    // various oscillators
    SinOsc s_2 => lpf;
    SawOsc saw_2 => lpf;
    TriOsc tri_2 => lpf;
    PulseOsc pul_2 => lpf;
    SqrOsc sqr_2 => lpf;
    // FM modulator and carrier
    TriOsc trictrl_2 => SinOsc sintri_2 => lpf;
    // interpret input as frequency modulation
    2 => sintri_2.sync;
    100  => trictrl_2.gain;
    
    // array of Oscs
    [ s_2, saw_2, tri_2, pul_2, sqr_2, trictrl_2 ] @=> Osc oscillators[];
    
    // set gains
    0.08 => s_2.gain;
    0.04 => saw_2.gain;
    0.04 => tri_2.gain;
    0.04 => pul_2.gain;
    0.04 => sqr_2.gain;
    0.04 => sintri_2.gain;
    
    // longer time loop
    spork ~ change_volume(notes, 1., 10.);
    now + 30::second => time later;
    while( now < later )
    { 
        // randomize
        Math.random2(0,7) => int select;
        // clamp (giving more weight to 5)
        if( select > 5 ) 5 => select;
        // generate new frequenc value
        Std.mtof( f[Math.random2( 0, 4 )] + 60 ) => float newnote;
        // set frequency
        newnote => oscillators[select].freq;
        // wait a bit
        0.5::second => now;
        // 10 times
        repeat(10)
        {
            Math.random2f( 0.2, 0.8 ) => trictrl_2.width;
            0.05::second => now;
        }
    }
    
    
    spork ~ change_volume(narrative, .5, 2.);
    4::second => now;
    
    // ------------------- MOVE THE NOTES AROUND ------------------- //
    
    float targets[oscillators.size()];
    
    change_freq(oscillators, targets, 10.);     randomize_freq(targets);
    spork ~ change_volume(notes, .7, 10.);
    change_freq(oscillators, targets, 10.);     randomize_freq(targets);
    spork ~ change_volume(notes, .9, 10.);
    change_freq(oscillators, targets, 10.);
    
    change_volume(notes, .4, 7.);
    change_volume(notes, 0., 4.);
}

fun void change_freq(Osc oscs[], float targets[], float duration) {
    <<< "change freq" >>>;
    now + duration::second => time later;
    
    float initials[0];
    for (Osc osc : oscs) {
        initials << osc.freq();
    }
    
    while( now < later )
    {
        // percentage (should go from 0 to 1)
        1 - (later-now)/(duration::second) => float progress;
        // for each sawtooth
        for( 0 => int i; i < oscs.size(); i++ ) {
            // update frequency by delta, towards target
            initials[i] + (targets[i]-initials[i])*progress => oscs[i].freq;
        }
        // advance time
        10::ms => now;
    } 
}

fun void randomize_freq(float targets[]) {
    [ 0, 2, 4, 7, 9 ] @=> int f[];
    for( 0 => int i; i < targets.size(); i++ ) {
        Std.mtof( f[Math.random2( 0, 4 )] + 60 ) => targets[i];
    }
}
play_narrative_notes();


// ------------------------- thunder and shit --------------------- //

fun void play_thunder() {
    <<< "play thunder" >>>;
    
    SndBuf thunder_buf1 => narrative;
    thunder_buf1.read("resources/thunder.wav");
    thunder_buf1.gain(0.8);
    15::second => now;
}

global SndBuf rain_buf;
spork ~ play_thunder();
narrative.gain(.8);
1.5::second => now;
spork ~ change_volume(notes, .2, 1.5);
5.5::second => now;

spork ~ play_thunder();
1.5::second => now;
spork ~ change_volume(notes, .5, 1.5);
3.5::second => now;

spork ~ play_thunder();
1.5::second => now;
change_volume(notes, .5, 8.);
spork ~ change_volume(rain_buf, 0.9, 7.);
10::second => now;
spork ~ change_volume(notes, 0., 10.);
spork ~ change_volume(rain_buf, 0.7, 10.);

1::eon => now;

